//==============================================================================
//	
//	Copyright (c) 2002-
//	Authors:
//	* Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford, formerly University of Birmingham)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

// includes
#include "PrismSparse.h"
#include <math.h>
#include <util.h>
#include <cudd.h>
#include <dd.h>
#include <odd.h>
#include <dv.h>
#include "sparse.h"
#include "PrismSparseGlob.h"
#include "jnipointer.h"
#include <new>

//------------------------------------------------------------------------------
// CUDA extensions

extern "C" int  mv_nd (int n, int nc, double *soln, int bound, bool min, bool instreward);
extern "C" void  datacpy_ndi (int n, int *row_starts, int *choice_starts, double *non_zeros, int nc, double *soln, unsigned int *cols);
extern "C" void datafree_ndi ();

//------------------------------------------------------------------------------

JNIEXPORT jlong __jlongpointer JNICALL Java_sparse_PrismSparse_PS_1NondetInstReward
(
JNIEnv *env,
jclass cls,
jlong __jlongpointer t,	// trans matrix
jlong __jlongpointer sr,	// state rewards
jlong __jlongpointer od,	// odd
jlong __jlongpointer rv,	// row vars
jint num_rvars,
jlong __jlongpointer cv,	// col vars
jint num_cvars,
jlong __jlongpointer ndv,	// nondet vars
jint num_ndvars,
jint bound,			// time bound
jboolean min,		// min or max probabilities (true = min, false = max)
jlong __jlongpointer in
)
{
	// cast function parameters
	DdNode *trans = jlong_to_DdNode(t);		// trans matrix
	DdNode *state_rewards = jlong_to_DdNode(sr);	// state rewards
	ODDNode *odd = jlong_to_ODDNode(od); 		// reachable states
	DdNode **rvars = jlong_to_DdNode_array(rv); 	// row vars
	DdNode **cvars = jlong_to_DdNode_array(cv); 	// col vars
	DdNode **ndvars = jlong_to_DdNode_array(ndv);	// nondet vars
	DdNode *init = jlong_to_DdNode(in);
	
	// model stats
	int n, nc;
	long nnz;
	// sparse matrix
	NDSparseMatrix *ndsm = NULL;
	// vectors
	double *soln = NULL, *soln2 = NULL, *tmpsoln = NULL;
	// timing stuff
	long start1, start2, start3, stop;
	double time_taken, time_for_setup, time_for_iters;
	// misc
	int i, j, k, l1, h1, l2, h2, iters;
	double d1, d2, kb, kbt;
	bool first;
	
	// exception handling around whole function
	try {
	
	// start clocks	
	start1 = start2 = util_cpu_time();
	
	// get number of states
	n = odd->eoff + odd->toff;
	
	// build sparse matrix (probs)
	PS_PrintToMainLog(env, "\nBuilding sparse matrix (transitions)... ");
	ndsm = build_nd_sparse_matrix(ddman, trans, rvars, cvars, num_rvars, ndvars, num_ndvars, odd);
	// get number of transitions/choices
	nnz = ndsm->nnz;
	nc = ndsm->nc;
	kb = (nnz*12.0+nc*4.0+n*4.0)/1024.0;
	kbt = kb;
	// print out info
	PS_PrintToMainLog(env, "[n=%d, nc=%d, nnz=%d, k=%d] ", n, nc, nnz, ndsm->k);
	PS_PrintMemoryToMainLog(env, "[", kb, "]\n");
	
	// create solution/iteration vectors
	// (solution is initialised to the state rewards)
	PS_PrintToMainLog(env, "Allocating iteration vectors... ");
	soln = mtbdd_to_double_vector(ddman, state_rewards, rvars, num_rvars, odd);
	soln2 = new double[n];
	kb = n*8.0/1024.0;
	kbt += 2*kb;
	PS_PrintMemoryToMainLog(env, "[2 x ", kb, "]\n");

	// print total memory usage
	PS_PrintMemoryToMainLog(env, "TOTAL: [", kbt, "]\n");

	// store local copies of stuff
	double *non_zeros = ndsm->non_zeros;
	unsigned char *row_counts = ndsm->row_counts;
	int *row_starts = (int *)ndsm->row_counts;
	unsigned char *choice_counts = ndsm->choice_counts;
	int *choice_starts = (int *)ndsm->choice_counts;
	bool use_counts = ndsm->use_counts;
	unsigned int *cols = ndsm->cols;
		
	// get setup time
	stop = util_cpu_time();
	time_for_setup = (double)(stop - start2)/1000;
	start2 = stop;

	if (use_gpu) {
		// Use row starts and choice starts
		if (use_counts) {
			row_starts = new int[n+1];
			row_starts[0] = 0;
			choice_starts = new int[nc];
			for (int i=1; i<n+1; i++) {
				row_starts[i] = row_starts[i-1] + row_counts[i-1];
			}
			choice_starts[0] = row_starts[0];
			for (int i=1; i<nc; i++) {
				choice_starts[i] = choice_starts[i-1] + choice_counts[i-1];
			}
		}
	}
	
	start2 = util_cpu_time();
	iters = 0;

	if (use_gpu) {
		PS_PrintToMainLog(env, "\nStarting GPU iterations...\n");
		datacpy_ndi (n,row_starts,choice_starts,non_zeros,nc,soln,cols);

		iters = mv_nd (n,nc,soln,bound,min,true);

		datafree_ndi ();
		fprintf (stdout, "GPU memory freed\n");
	}
	else {
		// start iterations
		PS_PrintToMainLog(env, "\nStarting iterations...\n");
	
		// note that we ignore max_iters as we know how any iterations _should_ be performed
		for (iters = 0; iters < bound; iters++) {
			// do matrix multiplication and min/max
			h1 = h2 = 0;
			// loop through states
			for (i = 0; i < n; i++) {
				d1 = 0.0;
				first = true;
				// get pointers to nondeterministic choices for state i
				if (!use_counts) { l1 = row_starts[i]; h1 = row_starts[i+1]; }
				else { l1 = h1; h1 += row_counts[i]; }
				// loop through those choices
				for (j = l1; j < h1; j++) {
					// compute the reward value for state i for this iteration
					// start with state reward for this state
					d2 = 0.0;
					// get pointers to transitions
					if (!use_counts) { l2 = choice_starts[j]; h2 = choice_starts[j+1]; }
					else { l2 = h2; h2 += choice_counts[j]; }
					// loop through transitions
					for (k = l2; k < h2; k++) {
						// add prob * corresponding reward from previous iteration
						d2 += non_zeros[k] * soln[cols[k]];
					}
					// see if this value is the min/max so far
					if (min) {
						if (first | (d2 < d1)) d1 = d2;
					} else {
						if (first | (d2 > d1)) d1 = d2;
					}
					first = false;
				}
				// set vector element
				// (if there were no choices from this state, reward is zero)
				soln2[i] = (h1 > l1) ? d1 : 0;
			}
		
			// prepare for next iteration
			tmpsoln = soln;
			soln = soln2;
			soln2 = tmpsoln;
		
		}
	}
	
	// stop clocks
	stop = util_cpu_time();
	time_for_iters = (double)(stop - start2)/1000;
	time_taken = (double)(stop - start1)/1000;
	
	// print iterations/timing info
	PS_PrintToMainLog(env, "\nIterative method: %d iterations in %.2f seconds (average %.6f, setup %.2f)\n", iters, time_taken, time_for_iters/iters, time_for_setup);
	
	// catch exceptions: register error, free memory
	} catch (std::bad_alloc e) {
		PS_SetErrorMessage("Out of memory");
		if (soln) delete[] soln;
		soln = 0;
	}
	
	// free memory
	if (ndsm) delete ndsm;
	if (soln2) delete[] soln2;
	//if (use_counts && use_gpu) { delete[] row_starts; delete[] choice_starts; }
		
	return ptr_to_jlong(soln);
}

//------------------------------------------------------------------------------
